package coursework;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import javafx.scene.control.Label;
import javafx.scene.control.TextField;
import javafx.scene.paint.Color;

public class Cage {

	private static int N;
	private int[] squareIds;
	private int value;
	private String operator;
	private Square[] gridSquares;
	private TextField[] gridNumbers;
	private Label[] cageLabels;
	
	public Cage(int[] squareIds, int value, String operator) {
		this.squareIds = squareIds;
		this.value = value;
		this.operator = operator;
		this.gridSquares = Main.getGridSquares();
		this.gridNumbers = Main.getGridNumbers();
		this.cageLabels = Main.getCageLabels();
		N = Main.getN();
		buildCage();
		setCageInfo();
	}
	
	private boolean plusOperator() {
		String gridValue;
		int cageVal = 0;
		for (int id : squareIds) {
			gridValue = gridNumbers[id].getText();
			if (!gridValue.equals("")) {
				cageVal += Integer.parseInt(gridValue);
			}
			else {
				return false;
			}
		}
		if (cageVal == value) {
			return false;
		}
		return true;
	}
	
	private boolean multiplyOperator() {
		String gridValue;
		int cageVal = 1;
		for (int id : squareIds) {
			gridValue = gridNumbers[id].getText();
			if (!gridValue.equals("")) {
				cageVal *= Integer.parseInt(gridValue);
			}
			else {
				return false;
			}
		}
		if (cageVal == value) {
			return false;
		}
		return true;
	}
	
	private boolean subtractOperator() {
		String gridValue;
		List<Integer> sortedCageValues = new ArrayList<Integer>();
		for (int id : squareIds) {
			gridValue = gridNumbers[id].getText();
			if (!gridValue.equals("")) {
				sortedCageValues.add(Integer.parseInt(gridValue));
			}
			else {
				return false;
			}
		}
		Collections.sort(sortedCageValues, Collections.reverseOrder());

		int cageVal = sortedCageValues.get(0);
		for (int i = 1; i < sortedCageValues.size(); i++) {
			cageVal -= sortedCageValues.get(i);
		}
		
		if (cageVal == value) {
			return false;
		}
		
		return true;
	}
	
	private boolean divideOperator() {
		String gridValue;
		List<Integer> sortedCageValues = new ArrayList<Integer>();
		for (int id : squareIds) {
			gridValue = gridNumbers[id].getText();
			if (!gridValue.equals("")) {
				sortedCageValues.add(Integer.parseInt(gridValue));
			}
			else {
				return false;
			}
		}
		Collections.sort(sortedCageValues, Collections.reverseOrder());
		
		int cageVal = sortedCageValues.get(0);
		for (int i = 1; i < sortedCageValues.size(); i++) {
			cageVal /= sortedCageValues.get(i);
		}
		
		if (cageVal == value) {
			return false;
		}
		return true;
	}
	
	public boolean checkCageValues() {
		switch (operator) {
		case "+":
			return plusOperator();
		case "x":
			return multiplyOperator();
		case "-":
			return subtractOperator();
		case "":
			return divideOperator();
		default:
			return false;
		}
	}
	
	public void buildCage() {
		if (squareIds.length == 1)
			return;
		
		for (int idA : squareIds) {
			for (int idB : squareIds) {
				if (idA != idB) {
					int[] aCoord = idToCoord(idA);
					int[] bCoord = idToCoord(idB);
					
					int rowDiff = aCoord[1] - bCoord[1];
					int colDiff = aCoord[0] - bCoord[0];
					
					final double WIDTH = 2;
					
					if (rowDiff == -1 && colDiff == 0) {
						gridSquares[aCoord[0]*N + aCoord[1]].setRightStroke(WIDTH, Color.GREY);
					}
					else if (rowDiff == 1 && colDiff == 0) {
						gridSquares[aCoord[0]*N + aCoord[1]].setLeftStroke(WIDTH, Color.GREY);
					}
					else if (rowDiff == 0 && colDiff == -1) {
						gridSquares[aCoord[0]*N + aCoord[1]].setBottomStroke(WIDTH, Color.GREY);
					}
					else if (rowDiff == 0 && colDiff == 1) {
						gridSquares[aCoord[0]*N + aCoord[1]].setTopStroke(WIDTH, Color.GREY);
					}
				}
			}
		}
	}
	
	public void setCageInfo() {
		if (squareIds.length > 1)
			Arrays.sort(squareIds);
		int minVal = squareIds[0];
		cageLabels[minVal].setText(value + operator);
	}
	
	public static int[] idToCoord(int id) {
		int row = id / N;
		int col = id % N;
		return new int[] {row, col};
	}
	
	public int[] getCageIds() {
		return squareIds;
	}
	
	public int getCageValue() {
		return value;
	}
	
	public String getCageOperator() {
		return operator;
	}

}