package coursework;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import javafx.scene.control.TextField;
import javafx.scene.layout.VBox;

public class ConstraintsHandler {

	private static int N;
	private TextField[] gridNumbers = new TextField[N*N];
	private VBox[] gridBoxes = new VBox[N*N];
	private List<Cage> gridCages = new ArrayList<Cage>();
	private boolean[][] gridErrors = new boolean[N*N][3];
	private boolean constraintChecked = false;	
	
	public ConstraintsHandler() {}
	
	private void updateData() {
		gridNumbers = Main.getGridNumbers();
		gridBoxes = Main.getGridBoxes();
		gridCages = Main.getGridCages();
		gridErrors = Main.getGridErrors();
		N = Main.getN();
	}
	
	public boolean checked() {
		return constraintChecked;
	}
	
	public void checkConstraints(int col, int row, boolean showMistakes) {
		constraintChecked = true;
		updateData();
		columnConstraint(row, col);
		rowConstraint(row, col);
		cageConstraint(row, col);
		highlightCells(showMistakes);
	}

	private void rowConstraint(int row, int col) {
		if (!doesRowRepeat(row)) {
			clearRow(row);
		}
		else {
			highlightRow(row);
		}
	}
	
	private boolean doesRowRepeat(int row) {
		boolean repeat = false;
		HashSet<String> rowValues = new HashSet<String>();
		String number;
		for (int i = 0; i < N; i++) {
			number = gridNumbers[row*N + i].getText();
			if (rowValues.contains(number) && !number.equals("")) {
				repeat = true;
			}
			else {
				rowValues.add(number);
			}
		}
		return repeat;
	}
	
	private void columnConstraint(int row, int col) {	
		if (!doesColumnRepeat(col)) {
			clearColumn(col);
		}
		else {
			highlightColumn(col);
		}
	}
	
	private boolean doesColumnRepeat(int col) {
		boolean repeat = false;
		HashSet<String> colValues = new HashSet<String>();
		String number;
		for (int i = 0; i < N; i++) {
			number = gridNumbers[i*N + col].getText();
			if (colValues.contains(number) && !number.equals("")) {
				repeat = true;
			}
			else {
				colValues.add(number);
			}
		}
		return repeat;
	}
		
	private void cageConstraint(int row, int col) {
		if (gridCages.size() > 0) {
			Cage searchCage = null;
			for (Cage cage : gridCages) {
				for (int id : cage.getCageIds()) {
					if (id == (row*N + col)) {
						searchCage = cage;
						break;
					}
				}
			}
			
			boolean validCageValues = searchCage.checkCageValues();
			for (int id : searchCage.getCageIds()) {
				gridErrors[id][2] = validCageValues;
			}
		}
	}
	
	private void clearRow(int row) {
		for (int i = 0; i < N; i++) {
			gridErrors[row*N + i][0] = false;		
		}
	}
	
	private void clearColumn(int col) {
		for (int i = 0; i < N; i++) {
			gridErrors[i*N + col][1] = false;
		}
	}
	
	private void highlightRow(int row) {
		for (int i = 0; i < N; i++) {
			gridErrors[row*N + i][0] = true;
		}
	}
	
	private void highlightColumn(int col) {
		for (int i = 0; i < N; i++) {
			gridErrors[i*N + col][1] = true;
		}
	}
	
	public void highlightCells(boolean showMistakes) {
		for (int i = 0; i < N*N; i++) {
			if ((gridErrors[i][0] || gridErrors[i][1] || gridErrors[i][2]) && showMistakes) {
				gridBoxes[i].setStyle("-fx-background-color: rgba(255,0,0,0.5);");
			}
			else {
				gridBoxes[i].setStyle("-fx-background-color: rgba(255,0,0,0);");
			}
		}
	}
	
	public boolean checkWinCondition() {
		updateData();
		if (checkCellsFull() && checkConstraintErrors()) {
			return true;
		}
		return false;
	}
	
	private boolean checkCellsFull() {
		System.out.println("");
		for (TextField textField : gridNumbers) {
			if (textField.getText().equals("")) {
				return false;
			}
		}
		return true;
	}
	
	private boolean checkConstraintErrors() {
		for (boolean[] cellErrors : gridErrors) {
			for (boolean errorVal : cellErrors) {
				if (errorVal) {
					return false;
				}
			}
		}
		return true;
	}
	
}
