package coursework;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Stack;
import java.util.stream.Collectors;

import javafx.application.*;
import javafx.geometry.Insets;
import javafx.geometry.Pos;
import javafx.scene.Group;
import javafx.scene.Scene;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.CheckBox;
import javafx.scene.control.Label;
import javafx.scene.control.RadioButton;
import javafx.scene.control.TextArea;
import javafx.scene.control.TextField;
import javafx.scene.control.ToggleGroup;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.GridPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Pane;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.scene.shape.Rectangle;
import javafx.stage.FileChooser;
import javafx.stage.Modality;
import javafx.stage.Stage;

public class Main extends Application {
	
	private static int N;
	
	private Scene scene;
	private Stage stage;
	
	private static Square[] gridSquares;
	private static TextField[] gridNumbers;
	private static Label[] cageLabels;
	private static VBox[] gridBoxes;
	private static List<Cage> gridCages;
	private static boolean[][] gridErrors;
	
	private static boolean showMistakes = false;
	private static Pane previousPane = null;
	private static Square previousSquare = null;
	private static boolean clearing = false;
	private static boolean undoing = false;
	private int fontSize;
	private static ConstraintsHandler constraints;
	
	private static Stack<GameState> undoStack;
	private static Stack<GameState> redoStack;
	
	private Button undoButton;
	private Button redoButton;
	
	//////////////////////////////////////////////////////////////////////////////////
	public static int getN() {
		return N;
	}
	
	public static Square[] getGridSquares() {
		return gridSquares;
	}
	
	public static TextField[] getGridNumbers() {
		return gridNumbers;
	}
	
	public static Label[] getCageLabels() {
		return cageLabels;
	}
	
	public static VBox[] getGridBoxes() {
		return gridBoxes;
	}
	
	public static List<Cage> getGridCages() {
		return gridCages;
	}
	
	public static boolean[][] getGridErrors() {
		return gridErrors;
	}
	//
	////////////////////////////////////////////////////////////////////////////////
	private GridPane setupGrid(double width, double height) {
		GridPane gridPane = new GridPane();
		gridPane.setAlignment(Pos.CENTER);
		double value = width < height ? width : height;	
		double size = value / N;
		for (int i = 0; i < N; i++) {
			for (int j = 0; j < N; j++) {
				gridPane = addToGridPane(gridPane, size, i, j);
			}
		}
		
		return gridPane;
	}
	
	private GridPane addToGridPane(GridPane gridPane, double size, int i, int j) {
		Square square = setupGridSquare(size);
		TextField textField = setupGridTextField(size);
		HBox hbox = new HBox(textField);
		hbox.setAlignment(Pos.CENTER);
		VBox vbox = setupGridVBox(square, i, j);
		GridPane.setMargin(vbox, new Insets(2));
		Group group = square.getGroup();
		
		gridSquares[i*N + j] = square;
		gridNumbers[i*N + j] = textField;
		
		gridPane.add(group, j, i);
		gridPane.add(hbox, j, i);
		gridPane.add(vbox, j, i);
		
		textField.textProperty().addListener((observable, oldValue, newValue) -> {
			textFieldDataChange(i, j);
		});
		
		square.getRectangle().toFront();
		return gridPane;
	}
	
	private VBox setupGridVBox(Square square, int i, int j) {
		VBox vbox = new VBox();
		Label cageInfoLabel = new Label("");
		
		cageLabels[i*N + j] = cageInfoLabel;
		
		vbox.getChildren().addAll(cageInfoLabel);
		vbox.setStyle("-fx-border-width: 0px;");
		vbox.setPadding(new Insets(2,0,0,5));
		gridBoxes[i*N + j] = vbox;
		paneClickSquareEvent(i, j, square);

		return vbox;
	}
	
	private Square setupGridSquare(double size) {
		final double WIDTH = 3;
		Square square = new Square(size);
		square.setTopStroke(WIDTH, Color.BLACK);
		square.setBottomStroke(WIDTH, Color.BLACK);
		square.setLeftStroke(WIDTH, Color.BLACK);
		square.setRightStroke(WIDTH, Color.BLACK);
		square.resizeLines(size);
		return square;
	}
	
	private TextField setupGridTextField(double size) {
		TextField textField = new TextField();
		textField.setMinWidth(size/4);
		textField.setPrefSize(size*0.9, size*0.9);
		textField.setAlignment(Pos.CENTER);
		textField.setPadding(new Insets(5));
		textField.setStyle("-fx-focus-color: transparent; -fx-text-box-border: transparent; -fx-background-color: -fx-text-box-border, -fx-control-inner-background;");
		return textField;
	}

	private HBox setupTopHBox() {
		HBox topHBox = new HBox();
		topHBox.setAlignment(Pos.BOTTOM_LEFT);
		topHBox.setSpacing(6);
		
		Button loadGameFileButton = new Button("Load Game - File");
		loadGameFileButton.setPrefWidth(140);
		loadGameFileButton.setPrefHeight(30);
		loadGameFileButton.setStyle("-fx-font-size:14px;");
		
		Button loadGameTextInputButton = new Button("Load Game - Text");
		loadGameTextInputButton.setStyle("-fx-font-size:14px;");
		loadGameTextInputButton.setPrefWidth(loadGameFileButton.getPrefWidth());
		loadGameTextInputButton.setPrefHeight(loadGameFileButton.getPrefHeight());
		
		Button generateRandomGameButton = new Button("Generate\nRandom Game");
		generateRandomGameButton.setPrefWidth(loadGameFileButton.getPrefWidth() - 20);
		generateRandomGameButton.setPrefHeight((loadGameFileButton.getPrefHeight() * 2) + 5);
		generateRandomGameButton.setStyle("-fx-text-alignment:center; -fx-font-size:14px;");
		
		VBox leftVBox = new VBox();
		leftVBox.setSpacing(5);
		leftVBox.setAlignment(Pos.BOTTOM_CENTER);
		leftVBox.getChildren().addAll(loadGameFileButton, loadGameTextInputButton);
		
		Button changeFontSizeButton = new Button("Font\nSize");
		changeFontSizeButton.setStyle("-fx-font-size:14px; -fx-text-alignment: center;");
		changeFontSizeButton.setPrefHeight((loadGameFileButton.getPrefHeight() * 2) + 5);
		changeFontSizeButton.setPrefWidth(60);
		
		topHBox.getChildren().addAll(leftVBox, generateRandomGameButton, changeFontSizeButton);
		
		changeFontSizeButton.setOnAction(e -> {
			changeFontSizePopup();
		});
		
		loadGameFileButton.setOnAction(e -> {
			loadFileButtonClickEvent();
		});
		
		loadGameTextInputButton.setOnAction(e -> {
			loadTextButtonClickEvent();
		});
		
		generateRandomGameButton.setOnAction(e -> {
			generateRandomGameEvent();
		});
		
		return topHBox;
	}
	
	private void updateGameState(GameState gameState) {
		String[] gameValues = gameState.getGameState();
		for (int i = 0; i < gridNumbers.length; i++) {
			gridNumbers[i].setText(gameValues[i]);
		}
	}
	
	private VBox setupBottomHBox() {
		VBox vbox = new VBox();
		vbox.setAlignment(Pos.CENTER);
		vbox.setSpacing(10);
		
		HBox bottomHBox = new HBox();
		bottomHBox.setSpacing(10);
		bottomHBox.setAlignment(Pos.TOP_CENTER);
		
//		undoButton = new Button("", new ImageView(new Image("file:Images/undo_icon.png")));
//		redoButton = new Button("", new ImageView(new Image("file:Images/redo_icon.png")));
		undoButton = new Button("Undo");
		undoButton.setStyle("-fx-font-size:14px;");
		redoButton = new Button("Redo");
		redoButton.setStyle("-fx-font-size:14px;");
		Button clearButton = new Button("Clear");
		undoButton.setOnAction(e -> {
			undoButtonClickEvent();
		});
		redoButton.setOnAction(e -> {
			redoButtonClickEvent();
		});
		
		undoButton.setDisable(true);
		redoButton.setDisable(true);
		clearButton.setStyle("-fx-font-size:14px;");
		clearButton.setPrefWidth(100);
		clearButtonClickEvent(clearButton);
		
		bottomHBox.getChildren().addAll(undoButton, clearButton, redoButton);
		
		CheckBox showMistakesCheck = new CheckBox("Show mistakes");
		showMistakesCheck.setMinWidth(showMistakesCheck.getPrefWidth());
		showMistakesCheck.setStyle("-fx-font-size:14px;");
		if (showMistakes) {
			showMistakesCheck.setSelected(true);
		}
		
		showMistakesCheck.selectedProperty().addListener((obs, oldVal, newVal) -> {
			showMistakesCheckEvent(showMistakesCheck);
		});
		
		vbox.getChildren().addAll(bottomHBox, showMistakesCheck);
		
		return vbox;
	}
	
	private VBox setupNumbersVBox() {
		VBox vbox = new VBox();
//		Button balanceButtonTop = new Button("");
//		balanceButtonTop.setPrefWidth(111);
//		balanceButtonTop.setMinWidth(111);
//		balanceButtonTop.setVisible(false);
//		Button balanceButtonBottom = new Button("");
//		balanceButtonBottom.setPrefWidth(111);
//		balanceButtonBottom.setMinWidth(111);
//		balanceButtonBottom.setVisible(false);
		//vbox.getChildren().add(balanceButtonTop);
		for (int i = 1; i <= N; i++)  {
			Button button = new Button(i + "");
			final double WIDTH = 50;
			
			button.setPrefSize(WIDTH, WIDTH);
			button.setMaxSize(WIDTH, WIDTH);
			button.setStyle("-fx-font-size:"+14+"pt");
			button.setOnAction(e -> {
				numbersButtonClickEvent(button);
			});
			vbox.getChildren().add(button);
		}
		//vbox.getChildren().add(balanceButtonBottom);
		vbox.setSpacing(5);
		vbox.setAlignment(Pos.CENTER_LEFT);
		vbox.setPadding(new Insets(0));
		return vbox;
	}
	
	
	private void changeFontSizePopup() {
		Stage popupStage = new Stage();
		popupStage.initModality(Modality.APPLICATION_MODAL);
		popupStage.initOwner(stage);
		
		VBox vbox = new VBox();
		
		HBox hbox = new HBox();	
		hbox.getChildren().addAll(
				setupButton("Tiny", 10, 75, popupStage),
				setupButton("Small", 14, 85, popupStage),
				setupButton("Medium", 18, 95, popupStage),
				setupButton("Large", 22, 105, popupStage),
				setupButton("Huge", 26, 115, popupStage));
		hbox.setSpacing(10);
		hbox.setAlignment(Pos.CENTER);
		
		Label headingLabel = new Label("Select font size:");
		headingLabel.setStyle("-fx-font-size: 18px;");
		
		vbox.setAlignment(Pos.CENTER);
		vbox.getChildren().addAll(headingLabel, hbox);
		
		
		popupStage.setScene(new Scene(vbox, 550, 110));
		popupStage.setTitle("Change font size");
		popupStage.show();
	}
	
	private Button setupButton(String name, int size, int width, Stage popupStage) {
		Button button = new Button(name);
		button.setMinWidth(width);
		button.setPrefWidth(width);
		button.setOnAction(e -> {
			changeFontSizeEvent(size, popupStage);
		});
		if (fontSize == size) {
			button.setStyle("-fx-font-size:" + size + "px; -fx-background-color: #99ddff; -fx-border-color: grey; fx-border-width: 1px; -fx-border-radius:3px;");
		}
		else {
			button.setStyle("-fx-font-size:" + size + "px;");
		}
		return button;
	}
	
	private void createNewCage(String line) {
		String[] splitLine = line.split(" ");
		if (splitLine.length < 2) {
			Alert alert = new Alert(AlertType.WARNING);
			alert.setContentText("The puzzle you have attempted to load is not in the correct format, please try again.");
			alert.setTitle("Incorrect Format");
			alert.showAndWait();
			return;
		}
		
		int[] ids;
		int value;
		String operator;
		
		if (splitLine[1].contains(",")) {
			String[] shiftedIds = splitLine[1].split(",");
			ids = new int[shiftedIds.length];
			
			for (int i = 0; i < shiftedIds.length; i++) {
				ids[i] = Integer.parseInt(shiftedIds[i]) - 1;
			}
			
			if (splitLine[0].contains("÷") || splitLine[0].contains("/")) {
				splitLine[0] = splitLine[0].replace("÷", "");
			}
			
			value = Integer.parseInt(splitLine[0].substring(0, splitLine[0].length()- 1));
			operator = splitLine[0].substring(splitLine[0].length() - 1);
		}
		else {
			ids = new int[1];
			ids[0] = Integer.parseInt(splitLine[1]) - 1;
			value = Integer.parseInt(splitLine[0]);
			operator = "";
		}
		
		Cage cage = new Cage(ids, value, operator);
		gridCages.add(cage);
		
		System.out.print(value + operator + " ");
		for (int i = 0; i < ids.length; i++) {
			System.out.print(ids[i] + 1);
			if (i < ids.length - 1) {
				System.out.print(",");
			}
		}
		System.out.print("\n");
	}
	
	
	private void clearGridAlert() {
		clearing = true;
		Alert alert = new Alert(AlertType.CONFIRMATION);
		alert.setContentText("Are you sure you wish to clear the board?");
		alert.setTitle("Clearing Board.");
		alert.showAndWait().ifPresent(type -> {
			if (type == ButtonType.OK) {
				clearGridData();
				undoStack.push(new GameState(GameState.getCurrentGameState(gridNumbers)));
			}
		});;
		clearing = false;
	}
	
	private void clearEmptyGridAlert() {
		Alert alert = new Alert(AlertType.WARNING);
		alert.setContentText("Nothing to clear on the board.");
		alert.setTitle("Empty Board");
		alert.showAndWait();
	}
	
	private void clearGridData() {
		for (TextField textField : gridNumbers) {
			textField.setText("");
		}
		for (int i = 0; i < N*N; i++) {
			for (int j = 0; j < 3; j++) {
				gridErrors[i][j] = false;
			}
		}
	}
		
	///////////////////////////////////////////////////////////
	private void generateRandomGameEvent() {
		Stage popupStage = new Stage();
		popupStage.initModality(Modality.APPLICATION_MODAL);
		popupStage.initOwner(stage);
		
		VBox vbox = new VBox();
		vbox.setSpacing(5);
		vbox.setAlignment(Pos.TOP_CENTER);
		
		HBox hboxBottom = new HBox();
		hboxBottom.setSpacing(100);
		hboxBottom.setPadding(new Insets(10));
		
		TextField textField = new TextField();		
		
		Button generateButton = new Button("Generate");
		generateButton.setAlignment(Pos.TOP_RIGHT);
		
		Button cancelButton = new Button("Cancel");
		cancelButton.setAlignment(Pos.TOP_LEFT);
		cancelButton.setOnAction(e -> {
			popupStage.hide();
		});
		
		HBox hboxTop = new HBox(new Label("Size (NxN):"), textField);
		hboxTop.setSpacing(3);
		hboxTop.setAlignment(Pos.CENTER);
		
		HBox hboxMiddle = new HBox();
		ToggleGroup tg = new ToggleGroup();
		RadioButton easy = new RadioButton("Easy");
		easy.setToggleGroup(tg);
		RadioButton medium = new RadioButton("Medium");
		medium.setToggleGroup(tg);
		RadioButton hard = new RadioButton("Hard");
		hard.setToggleGroup(tg);
		hboxMiddle.setAlignment(Pos.CENTER);
		hboxMiddle.setSpacing(10);
		VBox radioVBox = new VBox(easy, medium, hard);
		radioVBox.setSpacing(5);
		hboxMiddle.getChildren().addAll(new Label("Difficulty:"), radioVBox);
		
		hboxBottom.getChildren().addAll(cancelButton, generateButton);
		hboxBottom.setAlignment(Pos.CENTER);
		vbox.getChildren().addAll(hboxTop, hboxMiddle, hboxBottom);
		vbox.setAlignment(Pos.CENTER);
		
		generateButton.setOnAction(e -> {
			int textFieldValue = - 1;
			try {
				textFieldValue = Integer.parseInt(textField.getText());
			}
			catch (NumberFormatException ex) {
				Alert alert = new Alert(AlertType.WARNING);
				alert.setContentText("The value for the board size to generate a puzzel is not valid, please try again.");
				alert.setTitle("File Not Found");
				alert.showAndWait();
			}
			if (textFieldValue > 0 && textFieldValue <= 10) {
				N = textFieldValue;
				popupStage.close();
				resetGrid();
				int difficulty;
				if (easy.isSelected()) {
					difficulty = 1;
				}
				else if (medium.isSelected()) {
					difficulty = 2;
				}
				else if (hard.isSelected()) {
					difficulty = 3;
				}
				else {
					difficulty = 2;
				}
				RandomGenerator rg = new RandomGenerator(N, difficulty);
				gridCages = rg.generate();
			}
			if (textFieldValue > 10) {
				Alert alert = new Alert(AlertType.INFORMATION);
				alert.setContentText("The maximum size is 10.");
				alert.setTitle("Size Too Large");
				alert.showAndWait();
			}
		});
		popupStage.setMinHeight(200);
		popupStage.setMinWidth(255);
		popupStage.setScene(new Scene(vbox, 240, 160));
		popupStage.show();
	}
	
	private void loadFileButtonClickEvent() {
		try {
			FileChooser fileChooser = new FileChooser();
			fileChooser.getExtensionFilters().add(new FileChooser.ExtensionFilter("Text Files", "*.txt"));
			File selectedFile = fileChooser.showOpenDialog(stage);
			BufferedReader file = new BufferedReader(new FileReader(selectedFile));
			
			
			List<String> fileLines = file.lines().collect(Collectors.toList());
			int max = -1;
			for (String line : fileLines) {
				if (line.split(" ")[1].contains(",")) {
					String[] rightHalfLine = line.split(" ")[1].split(",");
					int[] lineIds = new int[rightHalfLine.length];
					for (int i = 0; i < lineIds.length; i++) {
						lineIds[i] = Integer.parseInt(rightHalfLine[i]);
						if (lineIds[i] > max) {
							max = lineIds[i];
						}
					}
				}
				else {
					int value = Integer.parseInt(line.split(" ")[1]);
					if (value > max) {
						max = value;
					}
				}
			}
			
			if (Math.sqrt(max) != N && Math.sqrt(max) <= 10) {
				N = (int) Math.sqrt(max);
				resetGrid();
			}
			
			if (Math.sqrt(max) > 10) {
				Alert alert = new Alert(AlertType.INFORMATION);
				alert.setContentText("The maximum size is 10.");
				alert.setTitle("Size Too Large");
				alert.showAndWait();
			}
			else {
				for (String line : fileLines) {
					createNewCage(line);
				}
				file.close();
			}
		}
		catch (FileNotFoundException e1) {
			Alert alert = new Alert(AlertType.WARNING);
			alert.setContentText("The file you have selected cannot be found, please try again.");
			alert.setTitle("File Not Found");
			alert.showAndWait();
		}
		catch (IOException e2) {
			Alert alert = new Alert(AlertType.WARNING);
			alert.setContentText("An error has occured when loading the file, please try again.");
			alert.setTitle("File Loading Error");
			alert.showAndWait();
		}
		catch (NullPointerException e3) {
			Alert alert = new Alert(AlertType.WARNING);
			alert.setContentText("You have not selected a file, please try again.");
			alert.setTitle("No File Selected");
			alert.showAndWait();
		}
		catch (NumberFormatException e4) {
			Alert alert = new Alert(AlertType.WARNING);
			alert.setContentText("The file you have selected is not in the correct format of a Mathdoku puzzle, please try again.");
			alert.setTitle("Incorrect File Format");
			alert.showAndWait();
		}
		
	}
	
	private void loadTextButtonClickEvent() {
		Stage popupStage = new Stage();
		popupStage.initModality(Modality.APPLICATION_MODAL);
		popupStage.initOwner(stage);
		
		VBox vbox = new VBox();
		vbox.setPadding(new Insets(10));
		vbox.setSpacing(5);
		
		HBox hboxBottom = new HBox();
		hboxBottom.setSpacing(186);
		
		
		HBox hboxTop = new HBox();
		hboxTop.setSpacing(5);
		hboxTop.setAlignment(Pos.CENTER);
		
		TextArea textArea = new TextArea();
		Button cancelButton = new Button("Cancel");
		cancelButton.setAlignment(Pos.TOP_LEFT);
		cancelButton.setOnAction(e -> {
			popupStage.hide();
		});
		Button loadButton = new Button("Load");
		loadButton.setAlignment(Pos.TOP_RIGHT);
		loadButton.setOnAction(e -> {
			try {
				String[] lines = textArea.getText().split("\n");
				int max = -1;
				for (String line : lines) {
					if (line.split(" ")[1].contains(",")) {
						String[] rightHalfLine = line.split(" ")[1].split(",");
						int[] lineIds = new int[rightHalfLine.length];
						for (int i = 0; i < lineIds.length; i++) {
							lineIds[i] = Integer.parseInt(rightHalfLine[i]);
							if (lineIds[i] > max) {
								max = lineIds[i];
							}
						}
					}
					else {
						int value = Integer.parseInt(line.split(" ")[1]);
						if (value > max) {
							max = value;
						}
					}
				}
				
				if (Math.sqrt(max) != N && Math.sqrt(max) <= 10) {
					N = (int) Math.sqrt(max);
					resetGrid();
				}
				
				if (Math.sqrt(max) > 10) {
					Alert alert = new Alert(AlertType.INFORMATION);
					alert.setContentText("The maximum size is 10.");
					alert.setTitle("Size Too Large");
					alert.showAndWait();
				}
				else {
					for (String line : lines) {
						if (!line.equals(null) && !line.equals("")) {
							createNewCage(line);
						}
					}
					popupStage.hide();
				}
			}
			catch (NumberFormatException e4) {
				Alert alert = new Alert(AlertType.WARNING);
				alert.setContentText("The file you have selected is not in the correct format of a Mathdoku puzzle, please try again.");
				alert.setTitle("Incorrect File Format");
				alert.showAndWait();
				popupStage.hide();
			}
		});
		
		hboxBottom.getChildren().addAll(cancelButton, loadButton);
		vbox.getChildren().addAll(textArea, hboxBottom);
		
		popupStage.setScene(new Scene(vbox, 300, 200));
		popupStage.setMinHeight(220);
		popupStage.setMinWidth(300);
		popupStage.show();
	}
	
	private void showMistakesCheckEvent(CheckBox showMistakesCheck) {
		if (!constraints.checked()) {
			constraints.checkConstraints(0, 0, false);
		}
		showMistakes = showMistakesCheck.isSelected();
		constraints.highlightCells(showMistakes);
	}
	
	private void undoButtonClickEvent() {
		undoing = true;			
		
		redoStack.push(undoStack.pop());
		updateGameState(undoStack.peek());
		
		if (!redoStack.empty()) {
			redoButton.setDisable(false);
		}
		if (undoStack.size() == 1) {
			undoButton.setDisable(true);
		}
		undoing = false;
	}
	
	private void redoButtonClickEvent() {
		GameState gameState = redoStack.pop();
		updateGameState(gameState);
		
		if (!undoStack.empty()) {
			undoButton.setDisable(false);
		}
		if (redoStack.empty()) {
			redoButton.setDisable(true);
		}
	}
	
	private void numbersButtonClickEvent(Button button) {	
		if (previousSquare != null) {
			for (int i = 0; i < gridSquares.length; i++) {
				if (previousSquare.equals(gridSquares[i])) {
					gridNumbers[i].setText(button.getText());
				}
			}
		}
	}
	
	private void changeFontSizeEvent(int size, Stage popupStage) {
		fontSize = size;
		for (TextField textField : gridNumbers) {
			textField.setStyle("-fx-focus-color: transparent; -fx-text-box-border: transparent;"
					+ "-fx-background-color: -fx-text-box-border, -fx-control-inner-background;"
					+ "-fx-font-size:" + size*1.8 + ";");
		}
		for (Label label : cageLabels) {
			label.setStyle("-fx-font-size:" + size*0.9 + ";");
		}
		if (popupStage != null) {
			popupStage.close();
		}
	}
	
	private void stageWidthResizeEvent(Stage stage, GridPane gridPane) {
		stage.widthProperty().addListener((obs, oldVal, newVal) -> {
			if (newVal.doubleValue() - 0 < stage.getHeight())  {
				resizeGrid();
			}
		});
	}
	
	private void stageHeightResizeEvent(Stage stage, GridPane gridPane) {
		stage.heightProperty().addListener((obs, oldVal, newVal) -> {
			if (newVal.doubleValue() < stage.getWidth() - 0)  {
				resizeGrid();
			}
		});
	}
	
	public void resizeGrid() {
		double width = scene.getWidth() - 170;
		double height = scene.getHeight() - 170;
		double newValue = height < width ? height: width;
		for (int i = 0; i < N; i++) {
			for (int j = 0; j < N; j++) {
				Rectangle rectangle = gridSquares[i*N + j].getRectangle();
				rectangle.setWidth(newValue / N);
				rectangle.setHeight(newValue / N);
				gridSquares[i*N + j].resizeLines(newValue / N);
			}
		}
	}
	
	private void paneClickSquareEvent(int i, int j, Square square) {
		gridBoxes[i*N + j].addEventHandler(MouseEvent.MOUSE_CLICKED, e -> {
			if (previousPane != null) {
				int previousIndex = getIndexOf(previousPane);
				if ((gridErrors[previousIndex][0] || gridErrors[previousIndex][1] || gridErrors[previousIndex][2]) && showMistakes) {
					previousPane.setStyle("-fx-background-color: rgba(255,0,20,0.5); -fx-border-width: 0px;");
				}
				else {
					previousPane.setStyle("-fx-background-color: rgba(255,0,20,0); -fx-border-width: 0px;");
				}
			}
			if ((gridErrors[i*N + j][0] || gridErrors[i*N + j][1] || gridErrors[i*N + j][2]) && showMistakes) {
				gridBoxes[i*N + j].setStyle("-fx-border-color: #2309FB; -fx-border-width: 2px; -fx-background-color: rgba(255,0,0,0.3);");
			}
			else {
				gridBoxes[i*N + j].setStyle("-fx-border-color: #2309FB; -fx-border-width: 2px; -fx-background-color: rgba(255,0,0,0);");
			}
			
			previousPane = gridBoxes[i*N + j];
			previousSquare = square;
			for (TextField textField : gridNumbers) {
				textField.deselect();
			}
			gridNumbers[i*N + j].requestFocus();
		});
	}
	
	private int getIndexOf(Pane pane) {
		for (int i = 0; i < N*N; i++) {
			if (pane.equals(gridBoxes[i])) {
				return i;
			}
		}
		return -1;
	}
	
	private void clearButtonClickEvent(Button clearButton) {		
		clearButton.setOnAction(e -> {
			if (!lastOperationClear()) {
				clearGridAlert();
			}
			else {
				clearEmptyGridAlert();
			}
		});
	}
	
	private void textFieldDataChange(int i, int j) {		
		if (!clearing && !undoing) {
			
			boolean validInput = true;
			try {
				int textInput = Integer.parseInt(gridNumbers[i*N + j].getText());
				if (textInput < 1) {
					validInput = false;
					gridNumbers[i*N + j].setText("");
				}
				else if (textInput > N) {
					validInput = false;
					gridNumbers[i*N + j].setText(gridNumbers[i*N + j].getText().substring(0, 1));
				}
			}
			catch (NumberFormatException ex) {
				validInput = false;
				if (gridNumbers[i*N + j].getText().length() == 2) {
					gridNumbers[i*N + j].setText(gridNumbers[i*N + j].getText().substring(0, 1));
				}
				else {
					gridNumbers[i*N + j].setText("");
				}
			}

			if (validInput) {
				undoStack.push(new GameState(GameState.getCurrentGameState(gridNumbers)));
				if (undoButton.isDisabled()) {
					undoButton.setDisable(false);
				}
			}
		}
		constraints.checkConstraints(j, i, showMistakes);;
		if (constraints.checkWinCondition()) {
			Animation animation = new Animation();
			animation.go();
		}
	}
	
	private boolean lastOperationClear() {
		GameState gameState = undoStack.peek();
		boolean emptyGrid = true;
		for (String textFieldValue : gameState.getGameState()) {
			if (textFieldValue.length() > 0) {
				emptyGrid = false;
			}
		}
		return emptyGrid;
	}
	
	////////////////////////////////////////////////////////
	
	@Override
	public void start(Stage primaryStage) throws Exception {
		stage = primaryStage;
		
		GridPane gridPane = new GridPane();
		int size = N*80;
		int extraWidth = 170;
		int extraHeight = 170;
		
		scene = new Scene(gridPane, size, size);
		stage.setWidth(size + extraWidth);
		stage.setHeight(size + extraHeight);
		gridPane.setPadding(new Insets(20));
		
		HBox topHBox = setupTopHBox();
		VBox numbersVBox = setupNumbersVBox();
		VBox bottomHBox = setupBottomHBox();
		GridPane puzzleGrid = setupGrid(size, size);
		undoStack.push(new GameState(GameState.getCurrentGameState(gridNumbers)));
		
		fontSize = 18;
		changeFontSizeEvent(fontSize, null);
		
		gridPane.add(topHBox, 0, 0, 2, 1);
		gridPane.add(puzzleGrid, 0, 1);
		gridPane.add(numbersVBox, 1, 1);
		gridPane.add(bottomHBox, 0, 2);
		
		gridPane.setVgap(10);
		gridPane.setHgap(10);
		gridPane.setAlignment(Pos.CENTER);
			
		stageWidthResizeEvent(stage, puzzleGrid);
		stageHeightResizeEvent(stage, puzzleGrid);
		
		stage.setMinWidth(size + extraWidth + 50);
		stage.setMinHeight(size + extraHeight + 50);
		stage.setScene(scene);
		stage.setTitle("Mathdoku");
		stage.show();
		
	}
	
	private void resetGrid() {
		stage.close();
		initialiseVariables();
		GridPane gridPane = new GridPane();;
		
		int size = N*80;
		int extraWidth = 170;
		int extraHeight = 170;
		
		if (N*80 + extraHeight + 50 > 1000) {
			System.out.println("Here");
			size = (1000 - extraHeight);
		}
		
		Stage newStage = new Stage();
		Scene newScene = new Scene(gridPane, size + extraWidth, size + extraHeight);
		newStage.setWidth(size + extraWidth + 50);
		newStage.setHeight(size + extraHeight + 50);
		gridPane.setPadding(new Insets(20));
		
		HBox topHBox = setupTopHBox();
		VBox numbersVBox = setupNumbersVBox();
		VBox bottomHBox = setupBottomHBox();
		GridPane puzzleGrid = setupGrid(size, size);
		undoStack.push(new GameState(GameState.getCurrentGameState(gridNumbers)));

		fontSize = 18;
		changeFontSizeEvent(fontSize, null);
		
		gridPane.add(topHBox, 0, 0, 2, 1);
		gridPane.add(puzzleGrid, 0, 1);
		gridPane.add(numbersVBox, 1, 1);
		gridPane.add(bottomHBox, 0, 2);
		
		gridPane.setVgap(10);
		gridPane.setHgap(10);
		gridPane.setAlignment(Pos.CENTER);
		
		stageWidthResizeEvent(stage, puzzleGrid);
		stageHeightResizeEvent(stage, puzzleGrid);
		
		newStage.setMinWidth(size + extraWidth + 50);
		newStage.setMinHeight(size + extraHeight + 50);
		newStage.setMaxHeight(1000);
		newStage.setMaxWidth(1200);
		newStage.setScene(newScene);
		newStage.setTitle("Mathdoku");
		newStage.show();
		scene = newScene;
		stage = newStage;

	}
	
	private static void initialiseVariables() {
		gridSquares = new Square[N*N];
		gridNumbers = new TextField[N*N];
		cageLabels = new Label[N*N];
		gridBoxes = new VBox[N*N];
		gridCages = new ArrayList<Cage>();
		gridErrors = new boolean[N*N][3];
		previousPane = null;
		previousSquare = null;
		clearing = false;
		undoing = false;
		undoStack = new Stack<GameState>();
		redoStack = new Stack<GameState>();
		constraints = new ConstraintsHandler();
	}
	
	public static void main(String[] args) {
		N = 4;
		initialiseVariables();
		launch(args);
	}
		
}