package coursework;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;

public class RandomGenerator {

	private int N;
	private int[][] grid;
	private int difficulty;	// 1 = easy | 2 = medium | 3 = hard	
	
	private List<Integer> uncagedGrids = new ArrayList<Integer>();
	private List<Integer> possibleMoves = new ArrayList<Integer>();
	private List<Cage> gridCages = new ArrayList<Cage>();
	
	public RandomGenerator(int N, int difficulty) {
		this.N = N;
		this.difficulty = difficulty;
		grid = new int[N][N];
	}
	
	public List<Cage> generate() {		
		generateGridValues();
		shuffleGridValues();
		generateCages();
		return gridCages;
	}
	
	private void generateCages() {
		Random random = new Random();
		addAllGrids();
		while (uncagedGrids.size() > 0) {
			int cageProbability = random.nextInt(100);
			int cageSize;
	
			
			switch (difficulty) {
			case 1:
				if (cageProbability < 15) {
					cageSize = 1;
				}
				else if (cageProbability < 80) {
					cageSize = 2;
				}
				else if (cageProbability < 95) {
					cageSize = 3;
				}
				else {
					cageSize = 4;
				}
				break;
			case 2:
				if (cageProbability < 50) {
					cageSize = 2;
				}
				else if (cageProbability < 85) {
					cageSize = 3;
				}
				else {
					cageSize = 4;
				}
				break;
			case 3:
				if (cageProbability < 45) {
					cageSize = 2;
				}
				else if (cageProbability < 75) {
					cageSize = 3;
				}
				else if (cageProbability < 97) {
					cageSize = 4;
				}
				else {
					cageSize = 5;
				}
				break;
				
			default:
				cageSize = 2;
			}
			
			
			if (uncagedGrids.size() <= cageSize) {
				cageSize = uncagedGrids.size();
			}
			if (cageSize == 0) {
				break;
			}
			
			int[] cageIds = new int[cageSize];
			int id = uncagedGrids.get(0);
			cageIds[0] = id;
			uncagedGrids.remove(0);
			for (int i = 0; i < cageSize - 1; i++) {
				int cornerGrid = isCornerGrid(id);
				int edgeGrid = isEdgeGrid(id);
				
				possibleMoves = new ArrayList<Integer>();
				possibleMoves.add(id + 1);
				possibleMoves.add(id - 1);
				possibleMoves.add(id + N);
				possibleMoves.add(id - N);
				
				if (cornerGrid != -1) {
					if (cornerGrid == 0) {
						topLeftCorner(id);
					}
					else if (cornerGrid == 1) {
						topRightCorner(id);
					}
					else if (cornerGrid == 2) {
						bottomRightCorner(id);
					}
					else if (cornerGrid == 3) {
						bottomLeftCorner(id);
					}
				}
				else if (edgeGrid != -1) {
					if (edgeGrid == 0) {
						leftEdge(id);
					}
					else if (edgeGrid == 1) {
						rightEdge(id);
					}
					else if (edgeGrid == 2) {
						topEdge(id);
					}
					else if (edgeGrid == 3) {
						bottomEdge(id);
					}
				}
				else {
					centre(id);
				}
				

				if (possibleMoves.size() >= 1) {
					id = possibleMoves.get(random.nextInt(possibleMoves.size()));
				}			
				cageIds[i + 1] = id;
				if (uncagedGrids.contains(id)) {
					uncagedGrids.remove(uncagedGrids.indexOf(id));
				}
			}
						
			int cageValue;
			String cageOperator;
			List<Integer> cageValues = new ArrayList<Integer>();
			HashSet<Integer> hashSetIds = new HashSet<Integer>();
			for (int cageId : cageIds) {
				hashSetIds.add(cageId);
			}
			
			if (hashSetIds.size() > 1) {
				for (int cageId : hashSetIds) {
					int row = cageId / N;
					int col = cageId % N;
					cageValues.add(grid[row][col]);
				}
				Collections.sort(cageValues, Collections.reverseOrder());
				
				Map<String, Integer> possibleValues = new HashMap<String, Integer>();
				
				int additionResult = 0;
				for (int value : cageValues) {
					additionResult += value;
				}
				possibleValues.put("+", additionResult);
				
				int subtractionResult = cageValues.get(0);
				for (int i = 1; i < cageValues.size(); i++) {
					subtractionResult -= cageValues.get(i);
				}
				if (subtractionResult > 0) {
					possibleValues.put("-", subtractionResult);
				}
				
				int multiplicationResult = 1;
				for (int value : cageValues) {
					multiplicationResult *= value;
				}
				possibleValues.put("x", multiplicationResult);
				
				float divisionResult = cageValues.get(0);
				for (int i = 1; i < cageValues.size(); i++) {
					divisionResult /= cageValues.get(i);
				}
				if (divisionResult == (int) divisionResult) {
					possibleValues.put("", (int) divisionResult);
				}
				
				
				List<String> keySet = new ArrayList<String>(possibleValues.keySet());
				cageOperator = keySet.get(random.nextInt(keySet.size()));
				cageValue = possibleValues.get(cageOperator);
			}
			else {
				int row = cageIds[0] / N;
				int col = cageIds[0] % N;
				cageValue = grid[row][col];
				cageOperator = "";
			}
			
			int[] finalCageIds = new int[hashSetIds.size()];
			int counter = 0;
			for (int cageId : hashSetIds) {
				finalCageIds[counter] = cageId;
				counter++;
			}
			
			gridCages.add(new Cage(cageIds, cageValue, cageOperator));
		}
	}

	
	private void centre(int id) {
		if (!uncagedGrids.contains(id + N))
			possibleMoves.remove(possibleMoves.indexOf(id + N));
		if (!uncagedGrids.contains(id - N))
			possibleMoves.remove(possibleMoves.indexOf(id - N));
		if (!uncagedGrids.contains(id + 1))
			possibleMoves.remove(possibleMoves.indexOf(id + 1));
		if (!uncagedGrids.contains(id - 1))
			possibleMoves.remove(possibleMoves.indexOf(id - 1));
	}
	
	private void topLeftCorner(int id) {
		possibleMoves.remove(possibleMoves.indexOf(id - 1));
		possibleMoves.remove(possibleMoves.indexOf(id - N));
		if (!uncagedGrids.contains(id + 1))
			possibleMoves.remove(possibleMoves.indexOf(id + 1));
		if (!uncagedGrids.contains(id + N))
			possibleMoves.remove(possibleMoves.indexOf(id + N));
	}
	
	private void topRightCorner(int id) {
		possibleMoves.remove(possibleMoves.indexOf(id - N));
		possibleMoves.remove(possibleMoves.indexOf(id + 1));
		if (!uncagedGrids.contains(id - 1))
			possibleMoves.remove(possibleMoves.indexOf(id - 1));
		if (!uncagedGrids.contains(id + N))
			possibleMoves.remove(possibleMoves.indexOf(id + N));
	}
	
	private void bottomRightCorner(int id) {
		possibleMoves.remove(possibleMoves.indexOf(id + N));
		possibleMoves.remove(possibleMoves.indexOf(id + 1));
		if (!uncagedGrids.contains(id - 1))
			possibleMoves.remove(possibleMoves.indexOf(id - 1));
		if (!uncagedGrids.contains(id - N))
			possibleMoves.remove(possibleMoves.indexOf(id - N));
	}
	
	private void bottomLeftCorner(int id) {
		possibleMoves.remove(possibleMoves.indexOf(id + N));
		possibleMoves.remove(possibleMoves.indexOf(id - 1));
		if (!uncagedGrids.contains(id + 1))
			possibleMoves.remove(possibleMoves.indexOf(id + 1));
		if (!uncagedGrids.contains(id - N))
			possibleMoves.remove(possibleMoves.indexOf(id - N));
	}
	
	private void leftEdge(int id) {
		possibleMoves.remove(possibleMoves.indexOf(id - 1));
		if (!uncagedGrids.contains(id + N))
			possibleMoves.remove(possibleMoves.indexOf(id + N));
		if (!uncagedGrids.contains(id - N))
			possibleMoves.remove(possibleMoves.indexOf(id - N));
		if (!uncagedGrids.contains(id + 1))
			possibleMoves.remove(possibleMoves.indexOf(id + 1));
	}
	
	private void topEdge(int id) {
		possibleMoves.remove(possibleMoves.indexOf(id - N));
		if (!uncagedGrids.contains(id + N))
			possibleMoves.remove(possibleMoves.indexOf(id + N));
		if (!uncagedGrids.contains(id + 1))
			possibleMoves.remove(possibleMoves.indexOf(id + 1));
		if (!uncagedGrids.contains(id - 1))
			possibleMoves.remove(possibleMoves.indexOf(id - 1));
	}
	
	private void rightEdge(int id) {
		possibleMoves.remove(possibleMoves.indexOf(id + 1));
		if (!uncagedGrids.contains(id - N))
			possibleMoves.remove(possibleMoves.indexOf(id - N));
		if (!uncagedGrids.contains(id + N))
			possibleMoves.remove(possibleMoves.indexOf(id + N));
		if (!uncagedGrids.contains(id - 1))
			possibleMoves.remove(possibleMoves.indexOf(id - 1));
	}
	
	private void bottomEdge(int id) {
		possibleMoves.remove(possibleMoves.indexOf(id + N));
		if (!uncagedGrids.contains(id - N))
			possibleMoves.remove(possibleMoves.indexOf(id - N));
		if (!uncagedGrids.contains(id - 1))
			possibleMoves.remove(possibleMoves.indexOf(id - 1));
		if (!uncagedGrids.contains(id + 1))
			possibleMoves.remove(possibleMoves.indexOf(id + 1));
	}
	
	private void addAllGrids() {
		for (int i = 0; i < N*N; i++) {
			uncagedGrids.add(i);
		}
	}
	
	private int isEdgeGrid(int gridId) {
		if (gridId % N == 0)			// left edge
			return 0;
		if (gridId % N == N - 1)		// right edge
			return 1;
		if ((int) gridId / N == 0)		// top edge
			return 2;
		if ((int) gridId / N == N - 1) 	// bottom edge
			return 3;
		return -1;		// not an edge grid
	}
	
	private int isCornerGrid(int gridId) {
		if ((int) gridId / N == 0     && gridId % N == 0) 		// top left corner
			return 0;
		if ((int) gridId / N == N - 1 && gridId % N == N - 1)	// bottom right corner
			return 2;
		if ((int) gridId / N == 0     && gridId % N == N - 1)	// top right corner
			return 1;
		if ((int) gridId / N == N - 1 && gridId % N == 0)		// bottom left corner
			return 3;
		return -1;		// not a corner
	}
	
	private void generateGridValues() {
		int counter = 1;
		for (int i = 0; i < N; i++) {
			for (int j = 0; j < N; j++) {
				grid[i][j] = counter;
				counter++;
				if (counter == N + 1) {
					counter = 1;
				}
			}
			counter--;
			if (counter == 0) {
				counter = N;
			}
		}
	}
	
	private void shuffleGridValues() {
		Random random = new Random();
		for (int i = 0; i < 15; i++) {
			int[] colsSwap = new int[2];
			int[] rowsSwap = new int[2];
			if (colsSwap[0] == colsSwap[1]) {
				colsSwap = generateValues(random);
			}
			if (rowsSwap[0] == rowsSwap[1]) {
				rowsSwap = generateValues(random);
			}
			
			swapColumns(colsSwap[0], colsSwap[1]);
			swapRows(rowsSwap[0], rowsSwap[1]);
			
			if (i == 7) {
				transpose();
			}
		}
	}
	
	private int[] generateValues(Random random) {
		return new int[] { random.nextInt(N), random.nextInt(N) };
	}
	
	private void swapColumns(int colA, int colB) {
		int[][] columns = new int[N][2];
		for (int i = 0; i < N; i++) {
			columns[i][0] = grid[i][colA];
			columns[i][1] = grid[i][colB];
		}
		for (int i = 0; i < N; i++) {
			grid[i][colA] = columns[i][1];
			grid[i][colB] = columns[i][0];
		}
	}
	
	private void swapRows(int rowA, int rowB) {
		int[][] rows = new int[N][2];
		for (int i = 0; i < N; i++) {
			rows[i][0] = grid[rowA][i];
			rows[i][1] = grid[rowB][i];
		}
		for (int i = 0; i < N; i++) {
			grid[rowA][i] = rows[i][1];
			grid[rowB][i] = rows[i][0];
		}
	}
	
	private void transpose() {
		int[][] temp = new int[N][N];
		for (int i = 0; i < N; i++) {
			for (int j = 0; j < N; j++) {
				temp[j][i] = grid[i][j];
			}
		}
		grid = temp;
	}

}